#include <Windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#pragma warning( disable : 4996 ) // disable deprecated warning 
#include <strsafe.h>
#pragma warning( default : 4996 )
#include <windowsx.h>
#include <Uxtheme.h>
#include <Dwmapi.h>

//#define IMPLICIT_CHAIN 1

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
LPDIRECT3D9          g_pD3D = NULL;
LPDIRECT3DDEVICE9    g_pd3dDevice = NULL;
IDirect3DSwapChain9* g_pswapChain = NULL;

MARGINS gMargins = { 10, 10, 32, 10 }; // left right top bottom

//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitD3D( HWND hWnd )
{
    // Create the D3D object.
    if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof( d3dpp ) );

    d3dpp.BackBufferFormat = D3DFMT_A8R8G8B8;
    d3dpp.SwapEffect = D3DSWAPEFFECT_COPY;
    d3dpp.Windowed = TRUE;
    d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    d3dpp.hDeviceWindow = hWnd;

    // Create the D3DDevice
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
                                      hWnd,
                                      D3DCREATE_MIXED_VERTEXPROCESSING |
                                      D3DCREATE_MULTITHREADED |
                                      D3DCREATE_FPU_PRESERVE,
                                      &d3dpp,
                                      &g_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: SetupSwapChain
// Desc: swap chain creation & management for the window 
//-----------------------------------------------------------------------------
bool SetupSwapChain( HWND hWnd )
{
  if (g_pswapChain) {
    IDirect3DSurface9* backBuffer = NULL;
    g_pswapChain->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO,
                                &backBuffer);
    D3DSURFACE_DESC desc;
    backBuffer->GetDesc(&desc);

    RECT r;
    GetClientRect(hWnd, &r);
    
    if (desc.Width == (r.right - r.left) && desc.Height == (r.bottom - r.top))
    {
      // Reuse
      g_pd3dDevice->SetRenderTarget(0, backBuffer);
      return true;
    }
  }

  // Reset

  D3DPRESENT_PARAMETERS d3dpp;
  memset(&d3dpp, 0, sizeof(D3DPRESENT_PARAMETERS));

  d3dpp.BackBufferFormat = D3DFMT_A8R8G8B8;
  d3dpp.SwapEffect = D3DSWAPEFFECT_COPY;
  d3dpp.Windowed = TRUE;
  d3dpp.BackBufferCount = 1;
  d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
  d3dpp.hDeviceWindow = hWnd;
  
  if (g_pswapChain) {
    g_pswapChain->Release();
    g_pswapChain = NULL;
  }

  if (FAILED(g_pd3dDevice->CreateAdditionalSwapChain(&d3dpp, &g_pswapChain)))
    return false;

  IDirect3DSurface9* backBuffer = NULL;
  g_pswapChain->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO,
                              &backBuffer);
  g_pd3dDevice->SetRenderTarget(0, backBuffer);

  return true;
}

//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
VOID Cleanup()
{
    if (g_pswapChain)
        g_pswapChain->Release();

    if( g_pd3dDevice != NULL )
        g_pd3dDevice->Release();

    if( g_pD3D != NULL )
        g_pD3D->Release();
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID Render( HWND hWnd )
{
#ifndef IMPLICIT_CHAIN
    SetupSwapChain(hWnd);
#endif

    // Clear the backbuffer and the zbuffer
    g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET,
                         0xFFA000A0, 0, 0 );

    // Begin the scene
    if( SUCCEEDED( g_pd3dDevice->BeginScene() ) )
    {
        // End the scene
        g_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
#ifdef IMPLICIT_CHAIN
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
#else
    g_pswapChain->Present( NULL, NULL, NULL, NULL, 0 );
#endif
}


//-----------------------------------------------------------------------------
// Name: ClientMarginHitTestPoint
// Desc: mouse hit detection for window edges
//-----------------------------------------------------------------------------
DWORD ClientMarginHitTestPoint(HWND hWnd, int mx, int my)
{
  // Calculations are done in screen coords
  RECT winRect;
  GetWindowRect(hWnd, &winRect);

  DWORD testResult = HTCLIENT;

  bool top    = false;
  bool bottom = false;
  bool left   = false;
  bool right  = false;

  int sideMargin = 8;
  int titlebarMargin = 32;

  if (my >= winRect.top && my < (winRect.top + titlebarMargin))
    top = true;
  else if (my < winRect.bottom && my >= (winRect.bottom - sideMargin))
    bottom = true;

  if (mx >= winRect.left && mx < (winRect.left +
                                  (bottom ? (2*sideMargin) : sideMargin)))
    left = true;
  else if (mx < winRect.right && mx >= (winRect.right -
                                        (bottom ? (2*sideMargin) : sideMargin)))
    right = true;

  if (top) {
    testResult = HTCAPTION;
    if (left)
      testResult = HTTOPLEFT;
    else if (right)
      testResult = HTTOPRIGHT;
  } else if (bottom) {
    testResult = HTBOTTOM;
    if (left)
      testResult = HTBOTTOMLEFT;
    else if (right)
      testResult = HTBOTTOMRIGHT;
  } else {
    if (left)
      testResult = HTLEFT;
    if (right)
      testResult = HTRIGHT;
  }
  return testResult;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    LRESULT dwmHitResult;
    if (DwmDefWindowProc(hWnd, msg, wParam, lParam, &dwmHitResult))
      return dwmHitResult;

    switch( msg )
    {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage( 0 );
            return 0;
        case WM_PAINT:
            Render(hWnd);
            return 0;
        case WM_NCCALCSIZE:
            return 0;
        case WM_NCHITTEST:
            return ClientMarginHitTestPoint(hWnd, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
        case WM_SIZE:
            break;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI wWinMain( HINSTANCE hInst, HINSTANCE, LPWSTR, INT )
{
    UNREFERENCED_PARAMETER( hInst );

    // Register the window class
    WNDCLASSEX wc =
    {
        sizeof( WNDCLASSEX ), CS_CLASSDC, MsgProc, 0L, 0L,
        GetModuleHandle( NULL ), NULL, NULL, NULL, NULL,
        L"D3D Tutorial", NULL
    };
    RegisterClassEx( &wc );

    // Create the application's window
    HWND hWnd = CreateWindow( L"D3D Tutorial", L"D3D Tutorial 06: Meshes",
                              WS_OVERLAPPEDWINDOW, 100, 100, 300, 300,
                              NULL, NULL, wc.hInstance, NULL );

    // Initialize Direct3D
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    {
        DwmExtendFrameIntoClientArea(hWnd, &gMargins);
        DWMNCRENDERINGPOLICY policy = DWMNCRP_USEWINDOWSTYLE;
        DwmSetWindowAttribute(hWnd, DWMWA_NCRENDERING_POLICY, &policy, sizeof policy);

        // Show the window
        ShowWindow( hWnd, SW_SHOWDEFAULT );
        UpdateWindow( hWnd );

        // Enter the message loop
        MSG msg;
        ZeroMemory( &msg, sizeof( msg ) );
        while( msg.message != WM_QUIT )
        {
            if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
            {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
            else
                Render(hWnd);
        }
    }

    UnregisterClass( L"D3D Tutorial", wc.hInstance );
    return 0;
}



